from  __future__ import division

from .flownetwork import FlowNetwork
from .tools import FlowDict

import numpy as np
import networkx as nx

TMAX = 200
TOL = 10e-6
NTRY=10

class LinearFlowNetwork(FlowNetwork):
    # The linear Poiseullie flow in a network
    def steady_flows(self):
        """
        The fixed points are given by:
            \sum_j (p_j-p_i)
        """
        L = nx.laplacian_matrix(self).toarray()
        I = np.array([self.node[n]['input'] for n in self.nodes()])

        pressures = np.insert(np.linalg.solve(L[1:, 1:], I[1:]), 0, 0)
        node_indices = {node: idx for idx, node in enumerate(self.nodes())}
        flows = FlowDict({(u, v): (pressures[node_indices[u]] - pressures[node_indices[v]]) * dat.get(
            self.weight_attr, 1) for (u, v, dat) in self.edges(data=True)})
        return flows



from hypothesis import given, assume
import hypothesis.strategies as st

from nose.tools import *

from flownetpy import LinearFlowNetwork

import numpy as np
import networkx as nx


class TestCore:

    def setUp(self):
        self.K_stable = 10

        # setup a 2-node graph
        G = nx.Graph()
        G.add_edge(1, 2, weight=self.K_stable)
        inputs = {1: 1, 2: -1}
        self.two_node_net = LinearFlowNetwork(G, inputs, weight='weight')

        # setup a ring network with inputs -1 +1 -1 +1...
        self.ring_size = 8
        ring_graph = nx.cycle_graph(self.ring_size)
        for u, v in ring_graph.edges():
            ring_graph[u][v]['weight'] = self.K_stable
        inputs = {
            node: (node % 2 - 0.5) * 2 for node in np.arange(self.ring_size)}
        self.ring_net_even = LinearFlowNetwork(
            ring_graph, inputs, weight='weight')

        # setup a ring network with inputs +1 +1 +1 +1 -1 -1 -1 -1 ...
        inputs = {node: (int(node < self.ring_size / 2) - 0.5)
                  * 2 for node in np.arange(self.ring_size)}
        self.ring_net_odd = LinearFlowNetwork(
            ring_graph, inputs, weight='weight')

    # Test fixed points for simple networks
    @given(st.floats(min_value=0.0001, max_value=10))
    def test_2node_fixed_point(self, K):
        """Fixed point should be 1"""
        for u, v in self.two_node_net.edges():
            self.two_node_net[u][v]['weight'] = K

        fp_2node = self.two_node_net.steady_flows()
        assert_almost_equal(fp_2node[(1, 2)], 1, places=4)

    @given(st.floats(min_value=0.0001, max_value=10))
    def test_even_ring_fixed_point(self, K):
        """The flows should look like -0.5 0.5 -0.5 0.5..."""
        for u, v in self.ring_net_even.edges():
            self.ring_net_even[u][v]['weight'] = K

        fp_ring = self.ring_net_even.steady_flows()
        assert(np.allclose([flow - (u % 2 - 0.5)
                            for (u, v), flow in fp_ring.items()], 0, atol=1e-6))

    @given(st.floats(min_value=0.0001, max_value=10))
    def test_odd_ring_fixed_point(self, K):
        """The flows should look like 1 -1 1 -1 ..."""
        for u, v in self.ring_net_odd.edges():
            self.ring_net_odd[u][v]['weight'] = K

        nnodes = self.ring_net_odd.number_of_nodes()
        fp_ring = self.ring_net_odd.steady_flows()
        fp_ring_array = [fp_ring[(i, (i + 1) % nnodes)] for i in range(nnodes)]
        fp_exact_array = [-1, 0, 1, 2, 1, 0, -1, -2]

        assert(np.allclose(fp_ring_array, fp_exact_array, atol=1e-6))




from hypothesis import given, assume
import hypothesis.strategies as st


from nose.tools import *

from flownetpy import KuramotoNetwork
from flownetpy.kuramotonetwork import _mod_pi, _omega, _random_stableop_initguess
import numpy as np
import networkx as nx


class TestHelperfuncs:
    """ Test the helpedr functions"""
    @given(x=st.floats(min_value=-10000, max_value=10000))
    def test_mod_pi_range(self, x):
        assert(np.abs(_mod_pi(x) < np.pi))

    @given(x=st.lists(st.floats(min_value=-10000, max_value=10000), min_size = 1, max_size = 100))
    def test_mod_pi_sanity(self, x):
        assert(np.all(np.abs(_mod_pi(x) < np.pi)))

    @given(m=st.integers(min_value=-10000, max_value=10000))
    def test_mod_pi_even_pi(self, m):
        """_mod_pi(even_multiple_of_pi)==0"""
        assert_almost_equal(_mod_pi(2 * m * np.pi), 0)

    @given(m=st.integers(min_value=-10000, max_value=10000))
    def test_mod_pi_odd_pi(self, m):
        """_mod_pi(odd_multiple_of_pi)==+/- 1"""
        assert_almost_equal(
            np.abs(_mod_pi((2 * m + 1) * np.pi)), np.pi)

    @given(size=st.integers(min_value=3))
    def test_random_stableop_initguess(self, size):
        """this function should return angles with |successive difference| < pi/2 """
        arr = _random_stableop_initguess(size)
        assert(
            (np.abs(_mod_pi(arr - np.roll(arr, 1))) <= np.pi / 2).all())

    @given(delta=st.floats(min_value=-np.pi / 2, max_value=np.pi / 2), size=st.integers(min_value=4, max_value=100))
    def test_omega_equispaced(self, delta, size):
        """
        checks omega of a ring network  with equispaced thetas
        """
        assume(not np.allclose(((size - 1) * delta / np.pi - 1) % 2, 0))
        self.ring_net = nx.cycle_graph(size)
        nnodes = self.ring_net.number_of_nodes()
        thetas = np.arange(nnodes) * delta
        assert_almost_equal(_omega(self.ring_net, [np.arange(size)], thetas)[0],
                            ((nnodes - 1) * delta + _mod_pi((1 - nnodes) * delta)) / 2 / np.pi)


class TestCore:

    def setUp(self):
        self.K_stable = 10

        # setup a 2-node graph
        G = nx.Graph()
        G.add_edge(1, 2, weight=self.K_stable)
        inputs = {1: 1, 2: -1}
        self.two_node_net = KuramotoNetwork(G, inputs, weight='weight')

        # setup a ring network with inputs -1 +1 -1 +1...
        self.ring_size = 8
        ring_graph = nx.cycle_graph(self.ring_size)
        for u, v in ring_graph.edges():
            ring_graph[u][v]['weight'] = self.K_stable
        inputs = {
            node: (node % 2 - 0.5) * 2 for node in np.arange(self.ring_size)}
        self.ring_net_even = KuramotoNetwork(
            ring_graph, inputs, weight='weight')

        # setup a ring network with inputs +1 +1 +1 +1 -1 -1 -1 -1 ...
        inputs = {node: (int(node < self.ring_size / 2) - 0.5)
                  * 2 for node in np.arange(self.ring_size)}
        self.ring_net_odd = KuramotoNetwork(
            ring_graph, inputs, weight='weight')

    # Test fixed points for simple networks
    @given(st.floats(min_value=0.0001, max_value=0.1))
    def test_2node_fixed_point(self, dK):
        """Fixed point should be 1"""
        for u, v in self.two_node_net.edges():
            self.two_node_net[u][v]['weight'] = 1+dK

        fp_2node, data = self.two_node_net.steady_flows(
            initguess=np.array([0, 0]), extra_output = True)
        assert_almost_equal(fp_2node[(1, 2)], 1, places=4)

    @given(st.floats(min_value=0.0001, max_value=0.1))
    def test_even_ring_fixed_point(self, dK):
        """The flows should look like -0.5 0.5 -0.5 0.5..."""
        for u, v in self.ring_net_even.edges():
            self.ring_net_even[u][v]['weight'] = 0.5 + dK

        fp_ring, data = self.ring_net_even.steady_flows(initguess=np.zeros(
            self.ring_net_even.number_of_nodes()), extra_output = True)
        assert(np.allclose([flow - (u % 2 - 0.5)
                            for (u, v), flow in fp_ring.items()], 0, atol=1e-6))

    @given(st.floats(min_value=0.0001, max_value=0.1))
    def test_odd_ring_fixed_point(self, dK):
        """The flows should look like 1 -1 1 -1 ..."""
        for u, v in self.ring_net_odd.edges():
            self.ring_net_odd[u][v]['weight'] = self.ring_size/4 + dK

        nnodes = self.ring_net_odd.number_of_nodes()
        fp_ring, data = self.ring_net_odd.steady_flows(
            initguess=np.zeros(nnodes), extra_output = True)
        fp_ring_array = [fp_ring[(i, (i + 1) % nnodes)] for i in range(nnodes)]
        fp_exact_array = [-1, 0, 1, 2, 1, 0, -1, -2]

        assert(np.allclose(fp_ring_array, fp_exact_array, atol=1e-6))

    # Test that no fixed point below critical coupling
    @given(st.floats(min_value=0.0001, max_value=0.1))
    def test_2node_unstable(self, dK):
        for u, v in self.two_node_net.edges():
            self.two_node_net[u][v]['weight'] = 1 - dK

        fp, data = self.two_node_net.steady_flows(
            initguess=np.array([0, 0]), extra_output = True)
        assert_is_none(fp)

    @given(st.floats(min_value=0.0001, max_value=0.1))
    def test_even_ring_unstable(self, dK):
        for u, v in self.ring_net_even.edges():
            self.ring_net_even[u][v]['weight'] = 0.5 - dK

        fp, data = self.ring_net_even.steady_flows(extra_output = True)
        assert_is_none(fp)

    @given(st.floats(min_value=0.0001, max_value=0.1))
    def test_odd_ring_unstable(self, dK):
        for u, v in self.ring_net_odd.edges():
            self.ring_net_odd[u][v]['weight'] = self.ring_size/4 - dK

        fp, data = self.ring_net_odd.steady_flows(extra_output = True)
        assert_is_none(fp)




"""
A module for flow network simulations
"""

from __future__ import division, print_function
import networkx as nx
from numbers import Number

class FlowNetwork(nx.Graph):
    """
    A class to describe a flow network. 

    A flow network is a graph with two properties:
        1.  Some amount of flow (called input) enters/leaves through its each node.
        2.  Each edge transports some flow (called current). 

    How the input flows distribute themselves into currents 
    is determined by a :class:`flowmodel`. 
    """

    def __init__(self, graph, inputs, weight=None):
        """
        Parameters
        ----------
        graph: networkx graph.
            DiGraphs and MultiGraphs not supported.
        input: dictionary.
            nodes as keys and inputs as values; e.g:
            {node_1: 1, node_2: -1, ...}
        flowmodel: class:`flowmodel` module.
        """
        nx.Graph.__init__(self, graph)
        if isinstance(weight, Number):
            # assign uniform weight to all edges
            for u,v in self.edges():
                self[u][v]['weight'] = weight
            self.weight_attr = 'weight'
        elif isinstance(weight, str):
            # assume graph aupplied already has specified weights
            # only point self.weight_attr to the correct attribute
            self.weight_attr = weight
        else:
            raise ValueError("We do not understand the meaning of weight %r"%weight)

        if isinstance(inputs, dict):
            # then inputs is a dictionary
            for node in inputs.keys():
                self.node[node]['input'] = inputs[node]
        else:
            # Assume inputs is a list-like object with
            # the same order as graph.nodes()
            for node, inpt in zip(graph.nodes(), inputs):
                self.node[node]['input'] = inpt

    def steady_flows(self, **kwargs):
        """
        Returns the steady state flows.

        Keyword Args
        ------------
            Whatever is required by the flows method of the flowmodel. 

        Returns
        -------
        F: dictionary
           A set of steady state flows in the network in form of a dictionary; s.t.
           F[(u,v)] == flow from u to v
        infodict: dictionary
           Additional conditions that lead to these flows. 

        Note
        ----
        Depending on the flowmodel, the returned flows may not be unique.
        """
        raise NotImplementedError



from  __future__ import division

import numpy as np
import networkx as nx
from scipy.integrate import ode

from .tools import FlowDict

TMAX = 200
TOL = 10e-6
NTRY=10

def steady_flows(flownet, initguess=None, extra_output=True):
    """
    Computes the steady state flows. 

    Args:
        flownet: A FlowNetwork object
        initguess: Initial conditions
        extra_output: boolean

    Returns:
        A dictionary
            d = {edge1 : flow1, edge2 : flow2,...}
        If extra_output=True, returns another dictionary
            data = {initguess: the_initial_condition, 'thetas': steady_state_thetas, 'omega': winding_vector}
    """

    thetas, initguess = _try_find_fps(NTRY, flownet, initguess=initguess)

    if thetas is None:
        return None, {'initguess': initguess}

    node_indices = {node: idx for idx, node in enumerate(flownet.nodes())}
    flows = FlowDict({(u, v): np.sin(thetas[node_indices[u]] - thetas[node_indices[v]])*
        dat.get(flownet.weight_attr, 1) for (u, v, dat) in flownet.edges(data=True)})

    if extra_output:
        omega = _omega(flownet, nx.cycle_basis(flownet), thetas)
        return flows, {'initguess': initguess, 'thetas': thetas, 'omega': omega}
    else:
        return flows

def _try_find_fps(ntry, flownet, tmax=TMAX, tol=TOL, initguess=None):
    """
    Tries to find a fixed point of the Kuramoto network. 

    Args:
        ntry    : number of initial conditions that will be tried to reach a fixed point
        flownet : a FlowNetwork  object
        tmax    : integration time
        tol     : the odesolver ends when the variance of thetas  are less than tol
        initguess : initial condition. If specified, ntry doesn't have any effect

    Returns:
        (fixed point, initguess)

    Note:
        If no fixed point is found, returns (None, initguess)
    """

    dt = tmax / 1000

    if initguess is not None: # then use the specified initguess    
        sol = _evolve(flownet, np.arange(0, tmax, dt), initguess = initguess)
        if _has_converged(sol):
            return sol[-1], initguess
        else:
            return None, initguess
    
    for ntry in range(ntry): # otherwise try `ntry` random initguesses
        initguess = _random_stableop_initguess(flownet.number_of_nodes())
        sol = _evolve(flownet, np.arange(0, tmax, dt), initguess)
        if _has_converged(sol):
            return sol[-1], initguess

    return None, initguess


def _evolve(flownet, tarr, initguess=None):
    """
    Evolves the flow network from `initguess` by timesteps in `tarr`
    """
    M = nx.incidence_matrix(flownet, oriented=True).toarray()
    Mw = nx.incidence_matrix(flownet, oriented=True, weight=flownet.weight_attr).toarray()
    P = np.array([flownet.node[n]['input'] for n in flownet.nodes()])
    
    if initguess is None:
        initguess = _random_stableop_initguess(flownet.number_of_nodes())
    
    return odeint(_kuramoto_ode, initguess, t=tarr, args=(M, Mw, P))


def _has_converged(time_series, window_size=0):
    """
    Detects if the time series has converged, by
    checking the values during the window given by 'window_size'
    """

    if window_size == 0:  # The window over which time series must be constant
        window_size = time_series.shape[0]//10 

    return np.allclose(np.var(time_series[-window_size:, :], axis=0), 0)


def odeint(func, x0, t=None, args=None, jac = None):
    """
    Integrate an ode for time array t
    """
    if jac:
        r = ode(func, func).set_integrator('vode', method='bdf')
        r.set_initial_value(x0, t[0]).set_f_params(*args)
        r.set_jac_params(*args)
    else:
        r = ode(func).set_integrator('vode', method='bdf')
        r.set_initial_value(x0, t[0]).set_f_params(*args)

    res = np.zeros((t.size, x0.size))
    res[0,:] = x0

    for idx, tnow in enumerate(t[1:]):
        r.integrate(tnow)
        res[idx+1, :] = r.y
    return res


def _kuramoto_ode(t, th, M_I, M_I_w, P):
    """
    Args:
	    th: theta, array with size=nnodes
	    t: time
	    M_I: unweighted oriented incidence matrix
	    M_I_W: weighted oriented incidence matrix
	    P: Power production at each node
    """
    return P - np.dot(M_I_w, np.sin(np.dot(M_I.T, th)))

def _kuramoto_jacobian(t, th, M_I, M_I_w, P):
    return -np.dot(M_I_w, np.cos(np.dot(M_I.T, th)))

def _omega(graph, cycles, thetas):
    """
    Calculates the winding number:
        (\sum_{i,j \in cycle} asin(\theta_j-\theta_i))/2\pi
    """
    omegas = []
    node2idx = {node:idx for idx, node in enumerate(graph.nodes())}
    for cycle in cycles:
        omega = 0
        angles_cycle = np.array([thetas[node2idx[node]]\
                       for node in cycle])
        phasediffs_cycle = _mod_pi(angles_cycle - np.roll(angles_cycle, 1))


        omega = np.sum(phasediffs_cycle) / np.pi / 2
        omegas.append(omega)
    return omegas


def _random_stableop_initguess(size):
    """
    Args:
        size: int

    Returns:
        An array res of random numbers s.t. 
        |res[i]-res[i+1 % size]| < \pi/2
    """
    low = -np.pi / 2
    high = np.pi / 2
    
    initguess = np.cumsum(np.random.uniform(low=low, high=high, size=size - 1))
    initguess = np.insert(initguess, 0, 0)

    if np.abs(_mod_pi(initguess[-1])) > np.pi/2:
        return _random_stableop_initguess(size)
    else:
        return initguess


def _mod_pi(angle):
    """
    given an angle, returns an equivalent angle
    within the interval [-pi,pi]

    Warning:
        Result is not well defined for angle = pi*(2k+1)
    """
    rem = np.remainder(angle, 2 * np.pi)
    return np.where(rem<=np.pi, rem, rem-2*np.pi)



class FlowDict(dict):
    """
    This is a custom dictionary class with tuples as keys. 
    If self[(u, v)] does not exist, then self[(u, v)] is 
    initialized with -self[(v, u)]
    """
    def __missing__(self, key):
        u, v = key
        try:
            self.__setitem__(key, -self.get((v, u)))
            return self[key]
        except KeyError:
            raise



import hypothesis.strategies as st
from hypothesis import given
import numpy as np

from flownetpy.tools import FlowDict

_MAX=10000
_MIN=-10000


class TestFlowDict:
    @given(st.dictionaries(keys=st.frozensets(st.integers(min_value=_MIN, max_value=_MAX), min_size=2, max_size=2), values=st.floats(min_value=_MIN, max_value=_MAX)))
    def test_sanity(self, data):
        """
        tests that FlowDict[(v ,u)] = -FlowDict[(u, v)]
        """
        x=FlowDict({(u,v):val for (u,v), val in data.items()})
        assert(np.allclose([x[(u,v)] - x[(u,v)] for u,v in data.keys()],
              0))
        assert(np.allclose([x[(v,u)] + x[(u,v)] for u,v in data.keys()],
              0))
    
